#!/usr/bin/env python

import sys, re, os, glob, shutil, types

from scipy import *
from scipy import interpolate
from utils import DmftEnvironment

import numpy
if numpy.__version__ == '1.0.1':
    loadtxt = io.read_array
    savetxt = io.write_array


def loadtxtE(filename):
    if numpy.__version__ == '1.0.1':
        return io.read_array(filename)[:,:-2]
    else:
        return loadtxt(filename)
    
            
class IMP_CTQMC(object):
    """Wrapper for ctqmc
    """
    def __init__(self, Znuc, dire, params, Sigind, CF):
        # Environment
        self.env = DmftEnvironment()
        self.mpi_prefix = self.env.MPI
        
        # impurity directory
        self.dir = dire+'/'

        # create impurity directory if it does not yet exist
        if len(glob.glob(dire))==0 : os.mkdir( dire )
        
        # log-file is opened
        self.fh_info = open(self.dir + 'info.out', 'w')

        CoulombF = params['CoulombF'][0].strip("'").strip('"')
        
        if CoulombF in ['Full', 'Georges']:
            print >> self.fh_info, 'CoulombF is one of Full | Georges'


        self.Sigind = array(Sigind)
        self.CF = CF

        # find smallest sigind to shift siginds so they start from 1
        minsigind = min([s for s in self.Sigind.flat if s != 0])
        self.Sigind_shifted = where(self.Sigind != 0, self.Sigind-minsigind+1, 0)
        
        # Transformation between spheric harmonics and local coordinates needs to be
        # used in exact diagonalization of the atom.
        fh_T = open(self.dir + 'Trans.dat', 'w')
        print >> fh_T, len(Sigind), len(Sigind[0]), '#  size of Sigind and CF'
        print >> fh_T, '#---- Sigind follows'
        for row in self.Sigind_shifted:
            for elem in row:
                print >> fh_T, ("%3s " % elem),
            print >> fh_T
        print >> fh_T, '#---- CF follows'

        for row in CF:
            for elem in row:
                print >> fh_T, "%12.8f %12.8f " % (elem.real, elem.imag),
            print >> fh_T

        # parameters, which will be given to ctqmc and will be written to PARAMS
        self.sparams={'exe':'ctqmc', 'Sig':'Sig.out', 'Delta':'Delta.inp', 'cix':'actqmc.cix', 'nom':100, 'aom':10, 'sderiv': 0.02, 'Ntau':1000, 'SampleGtau':1000, 'Naver':100000000, 'Ncout':1000000, 'Nmax':1024, 'GlobalFlip':10000, 'Gf':'Gf.out'}

        # If old distribution of kinks exists, it is copied to imp directory
        status = glob.glob("status.*")
        if self.dir!='./':
            #print 'self.dir=', self.dir
            for fil in status:
                shutil.copy2(fil, self.dir+'/'+fil)
        

        self.PARAMS = 'PARAMS'
        
        # file from database which contains atomic information for this particular atom
        database_arg = self.env.ROOT + '/database/' + 'atom.'+str(Znuc)+'.py'
        execfile(database_arg)
        database = locals() # it seems python does not know about variables obtained by execfile
        

        # The following variables will be set by the following procedure:
        #   if the variable is defined in params, it takes its value from params otherwise it checks its definition in database
        vars = ['l', 'para', 'qOCA', 'kOCA', 'mOCA', 'Ex', 'Ep', 'J', 'cx', 'Eoca', 'qatom', 'n', 'Ekeep', 'Nmax', 'Ekeepc', 'Nmaxc', 'Ewindow', 'max_M_size', 'CoulombF', 'OCA_G', 'add_occupancy', 'CoulombF']

        
        self.spr={}
        for var in vars:
            if params.has_key(var):
                self.spr[var] = params[var][0]
            elif database.has_key(var):
                self.spr[var] = database[var]
            if params.has_key('atom_'+var):
                self.spr[var] = params['atom_'+var][0]
                
        # Some variables have different name in database and input file
        # spr['n'] -- params['nc']  -- database['n']
        if params.has_key('nc'):
            self.spr['n'] = params['nc'][0]
        else:
            self.spr['n'] = database['nc']
        
        if params.has_key('Ncentral'):
            self.spr['Ncentral'] = params['Ncentral'][0]
        else:
            self.spr['Ncentral'] = self.spr['n'][1:len(self.spr['n'])-1]
        
        #self.deg={}
        #for b in range(len(Impq)):
        #    if self.deg.has_key(Impq[b]): self.deg[Impq[b]]+=1
        #    else: self.deg[Impq[b]]=1
        
        #if type(Impq)!= list :  Impq = Impq.tolist()
        #self.Impq = Impq
        #print >> self.fh_info, 'Impq=', Impq
        
        # stores some of just defined variables
        self.l = self.spr['l']
        self.J = self.spr['J']
        self.cx = self.spr['cx']

    
    def _exe(self, params, extn, UpdateAtom, gbroad=0.0, kbroad=0.0, maxAc=200.):
        """ Executes the CTQMC impurity solver
        """

        # Reading impurity levels
        (Eimp,Olap,Edc) = loadtxtE(self.dir+'/Eimp.inp')
        # Reading Delta
        Delta_dat = loadtxt(self.dir+'/Delta.imp').transpose()
        om = Delta_dat[0]
        Delta = Delta_dat[1::2] + Delta_dat[2::2]*1j
        # subtracting Edc, because the ksum used (Hk+s_oo-Edc)
        Eimps = Eimp-Edc 
        params['Ed'] = [Eimps.tolist(), "# Impurity levels"]
        
        # Our choice for mu_QMC is the following:
        #   impurity level for 5/2 is E_{5/2}=E0-2*cx and for 7/2 is E_{7/2}=E0+3/2*cx
        #   Here we correct E_{5/2} for 2*cx, which is already counted in cix file
        #E0 = params['Ed'][0][0]+2*self.cx
        E0 = Eimps[0]+2*self.cx
        mu_QMC = -E0
        
        # Below exact diagonalization of the atom is run to produce 'cix.out' file
        if UpdateAtom or len(glob.glob(self.dir+'/'+self.sparams['cix']))==0 : # no cix file yet -> create it

            
            if (self.l == 3) :
                # executable for ED of the atom
                self.atom_exe = self.env.ROOT + '/atomh'
                # argument list is created to later run the executable for exact diagonalization of the atom
                atom_arg = '-ns %d -ne %d ' % (self.spr['n'][0], self.spr['n'][-1])
                for k,p in self.spr.items():
                    if type(p)==types.ListType:
                        atom_arg += '-'+k+' "'+str(p)+'" '
                    else:
                        atom_arg += '-'+k+' '+str(p)+' '
                # print to info file
                print >> self.fh_info, 'atom_arg=', atom_arg
                # runs ED for the atom
                stdin, stdout, stderr = os.popen3('cd '+ self.dir + '; ' + self.atom_exe +' ' +atom_arg+' > nohup.out 2>&1')
                print >> self.fh_info, stdout.read()
                print >> self.fh_info, stderr.read()

            elif (self.l==2) :
                # executable for ED of the atom
                self.atom_exe = self.env.ROOT + '/atom_d.py'
                
                if type(Eimps)!= list :  Eimps = Eimps.tolist()
                print 'Eimps=', Eimps, 'E0=', E0
                Eimps = (array(Eimps)-E0).tolist()
                print 'Eimps_new=', Eimps, ' muQMC=', mu_QMC
                
                atom_arg = ''
                for k,p in self.spr.items():
                    atom_arg += '\"'+k+'='+str(p)+'\" '
                atom_arg += ' \"Eimp='+str(Eimps)+'\" '
                
                print >> self.fh_info, 'atom_arg=', atom_arg
                print atom_arg
                
                # runs ED for the atom
                cmd = 'cd '+ self.dir + '; ' + self.atom_exe +' ' +atom_arg+' > nohup.out 2>&1'
                print '.... running ..', cmd
                stdin, stdout, stderr = os.popen3( cmd )
                print >> self.fh_info, stdout.read()
                print >> self.fh_info, stderr.read()

                # If new cix file is created, the eigenvectors are reordered and hence
                # previous ctqmc kinks can not be directly used. One would need to reorder
                # them. But this is not very simple. Rather remove the status files.
                for filename in glob.glob(self.dir+'/status.*') :
                    os.remove( filename )
                
            else:
                print >> self.fh_info, 'l=', self.l, ' not yet implemented!'
        
        
        self.fh_info.flush()
        
        params['mu'] = [mu_QMC, '# QMC chemical potential']
        
        # executable
        shutil.copy2(self.env.ROOT+'/'+params['exe'][0], self.dir+'/'+params['exe'][0])
        self.q_exe = './'+params['exe'][0]
        
        # Preparing params.dat file
        fp = open(self.dir+self.PARAMS, 'w')
        kparams = params.keys()
        for p in self.sparams.keys():
            if p not in kparams:  # if this parameter is overwritten by the input parameters, skip it
                fp.write(p + '  ' + str(self.sparams[p]) + '\n')
            
        for p in params.keys():
            print >> fp, "%s  %s " % (p, params[p][0]), '\t ', params[p][1]
        fp.close()

        # Preparing Delta.inp file
        #print 'Printing to ', self.dir+self.sparams['Delta']
        fp = open(self.dir+self.sparams['Delta'], 'w')
        # Checking that Delta is causal and not regular
        Delta=array(Delta)
        for im in range(shape(Delta)[0]):
            for ib in range(shape(Delta)[1]):
                if abs(Delta[im,ib])>maxAc*pi : Delta[im,ib] = -1e-10j
                if Delta[im,ib].imag>0 : Delta[im,ib] = Delta[im,ib].real-1e-10j

        # creating big mesh
        T=om[0]/pi
        m_max = round(om[-1]/(pi*T))
        lom = arange(1,m_max+1,2)*pi*T
        # interpolate on big mesh
        lDelta=[]

        print 'shape(Delta)=', shape(Delta)
        print 'shape(om)=', shape(om)
        print 'shape(lom)=', shape(lom)
        #print 'lom=', lom
        print 'ss=', shape(Delta[0].real)
        
        for ib in range(len(Delta)):
            tckr = interpolate.splrep(om, Delta[ib].real, s=0)
            tcki = interpolate.splrep(om, Delta[ib].imag, s=0)
            lDelta.append( interpolate.splev(lom, tckr) + interpolate.splev(lom, tcki)*1j )
            
        for im,ome in enumerate(lom):
            print >> fp, ome,
            for b in range(len(lDelta)):
                print >> fp, "%f  %f " % (lDelta[b][im].real, lDelta[b][im].imag),
            print >> fp
        fp.close()
        
        shutil.copy2(self.dir+self.sparams['Delta'], self.dir+self.sparams['Delta']+'.'+extn)
        
        # Below we execute ctqmc
        cmd = 'cd '+self.dir+'; '+self.mpi_prefix+' '+self.q_exe+' '+self.PARAMS+' > nohup_imp.out 2>&1 '
        print cmd
        print 'Running ---- ctqmc -----'
        stdin, stdout, stderr = os.popen3(cmd)
        print >> self.fh_info, stdout.read(), stderr.read()

        
    def HighFrequency(self, params, DCs, da, extn, wbroad=0.0, kbroad=0.0):
        """ Reads impurity output and prepares for further execution
          Output:
               Sig[b][iom]  -- DMFT dynamic self-energy which vanishes at infinity
               sinf[b]      -- DMFT self-energy at infinity
               Edc[b]       -- double counting using DMFT occupancies
        """

        if kbroad>0.0 or wbroad>0.0:
            bexe = self.env.ROOT + '/broad'
            shutil.move(self.dir+self.sparams['Sig'], self.dir+self.sparams['Sig']+'b')
            broad_cmd = 'cd '+self.dir+'; '+bexe+' -w '+str(wbroad)+' -k '+str(kbroad)+' '+self.sparams['Sig']+'b  > '+self.sparams['Sig']
            print broad_cmd
            stdin, stdout, stderr = os.popen3(broad_cmd)
            print stderr.read()
            
            
        # Reading Sig.out which was produced by ctqmc
        fS = open(self.dir + self.sparams['Sig'])
        # Sig.out contains nf, moment, ntot,...
        first_line = fS.next().strip()
        adat = first_line.split()
        mom=0
        TrSigmaG=0
        for par in adat: # setting nf from ctqmc output
            m = re.search('nf=', par) or re.search('TrSigmaG=', par) or re.search('mom=', par)
            if m is not None : exec(par)
            
        ntot = nf
        # Reading the rest of Sig.out
        om=[]
        Sigo=[]
        for line in fS:
            dat = line.split()
            om.append(float(dat[0]))
            del dat[0]
            sb=[]
            nbath = len(dat)/2
            for b in range(nbath):
                sr = float(dat[2*b])
                si = float(dat[2*b+1])
                if si>0.0: si=0.0  # Must be causal
                sb.append(sr + si*1j)
            Sigo.append(sb)            
        Sigo=array(Sigo)
        om=array(om)

        # self-energy at infinity
        sinf = copy(Sigo[-1][:]).real


        #print 'sinf=', sinf
        
        # subtracting value at infinity
        for im in range(len(Sigo)): Sigo[im] -= sinf
        #print 's_oo=', Sigo[-1]
        
        Sigo = Sigo.transpose()

        
        #da = 0
        #if DCs=='default' : da = DCs['a']
        # Here we compute self-energy(infinity) and 
        # double-counting using impurity occupancy
        Ud = params['U'][0]
        if type(self.J) is list:
            Jd = sum(self.J)/len(self.J)
        else:
            Jd = self.J
        
        # If we want to have U different than the U used in double counting
        # (useful for cluster-DMFT when U can not be simply added in ctqmc)
        if params.has_key('Ud'): Ud = params['Ud'][0]
        if params.has_key('Jd'): Jd  = params['Jd'][0]
        
        Edc0 = Ud*(nf-0.5-da) - Jd*(nf/2.-0.5)
        Phidc0 = Ud*0.5*nf*(nf-1.-2*da) - Jd*0.25*nf*(nf-2.)
        
        # If user fixes double counting by certain predescribed nf0
        # (fixn=True), we need to normalized actual nf[:] such that
        # sum(nf[:]) is equal to predefined nf0
        # Here we do not change Sigma_{infinity} but only Edc.
        # Sigma_{infinity} is determined by actual nf, while Edc is determined by modified nf.
        CoulombF = params['CoulombF'][0].strip("'").strip('"')
        

        # Reading old Edc and impurity levels
        (Eimp_old,Olap_old,Edc_old) = loadtxtE(self.dir+'/Eimp.inp')

        # Average Sigma_oo
        asinf=0.0
        nw=0
        Sigind = array(self.Sigind_shifted)
        
        print >> self.fh_info, 'Edc_old=', Edc_old
        print >> self.fh_info, 'Sigind=', Sigind
        
        for i in range(len(Sigind)):
            if Sigind[i,i]>0 and Edc_old[Sigind[i,i]-1]>-1000 :
                asinf += sinf[Sigind[i,i]-1]
                nw += 1
        asinf *= 1./nw
        print >>self.fh_info, '<s_oo>=', asinf
        print '<s_oo>=', asinf
        print >>self.fh_info, 'DCs==', DCs, DCs=='s_oo'
        
        if DCs=='fixn':
            nf0 = params['nf0'][0]
            if CoulombF in ['Full', 'Georges']:
                Edc0 = Ud*(nf0-0.5) - Jd*(nf0/2.-0.5)
                Phidc0 = Ud*0.5*nf0*(nf0-1.) - Jd*0.25*nf0*(nf0-2.)
            else:
                Edc0 = Ud*(nf0-0.5) - Jd*(nf0/2.-0.5)/4.
                Phidc0 = Ud*0.5*nf0*(nf0-1.) - Jd*0.25*nf0*(nf0-2.)/4
        elif DCs=='s_oo':
            Edc0 = asinf
            Phidc0 = 0.5*asinf*ntot
            

        Edc = Edc_old # we want to keep Edc<-1000 for orbitals which are projected out!
        for i in range(len(Edc)):
            if (Edc_old[i]>-1000.): # this is true for all orbitals which are kept!
                Edc[i] = Edc0
            else:                   # these orbitals are projected out -> we want to keep sinf=Edc for orbitals projected out!
                sinf[i] = Edc[i]
                
        if os.path.isfile(self.dir+'/../Uc.dat') and os.path.getsize(self.dir+'/../Uc.dat')>0:
            dEdc = eval(open(self.dir+'/../Uc.dat','r').readline()[1:])
            Edc = Edc + array(dEdc)
            print >>self.fh_info, 'Found Uc.dat! dEdc=', dEdc
            print 'Found Uc.dat! dEdc=', dEdc
            
        
            
        # Copying data at each iteration to follow the evolution
        shutil.copy2(self.dir+self.sparams['Sig'], self.dir+self.sparams['Sig']+'.'+extn)
        shutil.copy2(self.dir+self.sparams['Gf'], self.dir+self.sparams['Gf']+'.'+extn)
        shutil.copy2(self.dir+'nohup_imp.out', self.dir+'nohup_imp.out'+'.'+extn)
        shutil.copy2(self.dir+'Probability.dat', self.dir+'Probability.dat'+'.'+extn)
        
        # # Reading Gf.out to see how good is doping
        # # For computing n_imp we need to subtract 1/(iom-E0)
        # E0 = array(params['Ed'][0])+array(sinf)
        # 
        # #fGo = io.read_array(self.dir+self.sparams['Gf'], columns=(0,-1), lines=(1,-1))
        # fGo = loadtxt(self.dir+self.sparams['Gf'])
        # nfb=zeros(nbath)
        # for ig in fGo:
        #     ome = ig[0]
        #     for b in range(nbath):
        #         nfb[b] += ig[2*b+1] - 1/(ome*1j-E0[b])
        # 
        # T=om[0]/pi
        # dn = [ferm(E0[b]/T) for b in range(nbath)]
        # n_imp = nfb*2*T + array(dn)
        # 
        # # occupation vector from OCA solver
        # # OCA gives only few numbers and we need to create vector - matrix with this
        # Sigind = array(self.Sigind)
        # size = 2*(2*self.l+1)
        # if len(Sigind)<size:
        #     SigindN = zeros((size,size), dtype=int)
        #     SigindN[:size/2,:size/2] = Sigind
        #     SigindN[size/2:,size/2:] = Sigind
        #     Sigind = SigindN
        # 
        # mx = max(map(max,Sigind))
        # deg = zeros(mx, dtype=float)

                
        # 
        # occj = zeros((size,size), dtype=complex, order='F')
        # nt_imp=0
        # for i in range(len(Sigind)):
        #     nt_imp += n_imp[Sigind[i,i]-1]/deg[Sigind[i,i]-1]
        #     for j in range(len(Sigind)):
        #         if Sigind[i,j]>0:
        #             occj[i,j] = n_imp[Sigind[i,j]-1]/deg[Sigind[i,j]-1]
        # 
        #print 'nimp,gnimp=', ntot, nt_imp
        #print >> self.fh_info, 'nimp,gnimp=', ntot, nt_imp
        
        print >> self.fh_info, 'nimp, TrSigmaG=', ntot, TrSigmaG
        
        self.fh_info.flush()
        
        soon=0 # sum_i s_oo(i) * n(i)
        if len(mom)>0:
            for i in range(len(sinf)):
                if Edc[i]>-1000: soon += sinf[i]*mom[i]

        Ry2eV = 13.60569193
        Epotential = TrSigmaG
        HartreeCorrection0 = 0.5*soon       # Full hartree term
        HartreeCorrection1 = 0.5*asinf*ntot # Rotatinally invariant (local) hartree term
        fE = open(self.dir+'/Eorb.dat', 'w')
        print >> fE, '  Tr(Sigma*G)/2=', Epotential
        print >> fE, '  <s_oo>=', asinf
        print >> fE, '  sum_i Sigma_oo(i)*n(i)/2=', HartreeCorrection0
        print >> fE, '  <Sigma_oo>*n/2=', HartreeCorrection1
        print >> fE, ':EORB ', (Epotential-HartreeCorrection1)/Ry2eV
        fE.close()
        
        fE = open(self.dir+'/sig.out', 'w')
        print >> fE, '# s_oo=', sinf.tolist()
        print >> fE, '# Edc=', Edc.tolist()
        #savetxt(fE, Sigo.transpose())
        
        for iom in range(len(om)):
            print >> fE, ("%20.15f " % om[iom]),
            for b in range(len(Sigo)):
                print >> fE, ("%20.15f %20.15f  " % (Sigo[b,iom].real+sinf[b]-Edc[b], Sigo[b,iom].imag)),
                #print >> fE, ("%20.15f %20.15f  " % (Sigo[b,iom].real, Sigo[b,iom].imag)),
            print >> fE
        fE.close()
        
        return ntot

    def non_exe(self, params, extn, UpdateAtom, gbroad=0.0, kbroad=0.0, maxAc=200.):
        """ Executes the CTQMC impurity solver
        """

        # Reading impurity levels
        (Eimp,Olap,Edc) = loadtxtE(self.dir+'/Eimp.inp')
        # Reading Delta
        Delta_dat = loadtxt(self.dir+'/Delta.imp').transpose()
        om = Delta_dat[0]
        Delta = Delta_dat[1::2] + Delta_dat[2::2]*1j
        # subtracting Edc, because the ksum used (Hk+s_oo-Edc)
        Eimps = Eimp-Edc 
        params['Ed'] = [Eimps.tolist(), "# Impurity levels"]
        
        # Our choice for mu_QMC is the following:
        #   impurity level for 5/2 is E_{5/2}=E0-2*cx and for 7/2 is E_{7/2}=E0+3/2*cx
        #   Here we correct E_{5/2} for 2*cx, which is already counted in cix file
        #E0 = params['Ed'][0][0]+2*self.cx
        E0 = Eimps[0]+2*self.cx
        mu_QMC = -E0

def ferm(x):
    """Fermi function for T=1"""
    if (x>100): return 0
    if (x<-100): return 1
    return 1/(exp(x)+1)


if __name__ == '__main__':

    iparams0={"exe"                : ["ctqmc"             , "# Name of the executable"],
          "U"                  : [4.                  , "# Coulomb repulsion (F0)"],
          "beta"               : [100.                , "# Inverse temperature"],
          "Nmax"               : [1000                , "# Maximum perturbation order allowed"],
          "M"                  : [20000000            , "# Total number of Monte Carlo steps"],
          "Ncout"              : [1000000             , "# How often to print out info"],
          "Naver"              : [100000000           , "# How often to print out debug info"],
          "nf0"                : [2.3                 , "# Double counting parameter"],
          "nc"                 : [[1, 2, 3, 4]           , "# Impurity occupancies"],
          "nom"                : [100                 , "# Number of Matsubara frequency points sampled"],
          "aom"                : [30                  , "# Number of frequency points used to determin the value of sigma at nom"],
          "sderiv"             : [0.02                , "# Maximum derivative mismatch accepted for tail concatenation"],
          "Ntau"               : [1000                , "# Number of imaginary time points (only for debugging)"],
          "SampleGtau"         : [1000                , "# How often to update G(tau)"],
          "GlobalFlip"         : [10000               , "# How often to try a global flip"],
          "tsample"            : [10                  , "# How often to record measurements"],
          "warmup"             : [1000000             , "# Warmup number of QMC steps"],
          "CleanUpdate"        : [100000              , "# How often to make clean update"],
          "minM"               : [1e-10               , "# The smallest allowed value for the atomic trace"],
          "minD"               : [1e-10               , "# The smallest allowed value for the determinant"],
          "Ncorrect"           : [-1                  , "# Which baths should not be corrected"],
          "PChangeOrder"       : [0.9                 , "# Ratio between trial steps: add-remove-a-kink / move-a-kink"],
          "ChooseRandom"       : [1.0                 , "# How often to choose kink in random bath and how often in those with less kinks"],
          "CoulombF"           : ["'Full'"            , "# Full Coulomb repulsion"],
          "OCA_G"              : [False               , "# No OCA diagrams being computed - for speed"],
    }


    fi = open('Trans.dat', 'r')
    dim = int(fi.next().split()[0])
    fi.next()
    Sigind=[]
    for i in range(dim):
        Sigind.append( map(int,fi.next().split()[:dim]) )
    Sigind = array( Sigind )
    fi.next()
    CF=[]
    for i in range(dim):
        dat = map(float,fi.next().split()[:2*dim])
        CF.append( array(dat[::2])+array(dat[1::2])*1j )
    CF = array( CF )
    imp = IMP_CTQMC(26,'.', iparams0, Sigind, CF)
    imp.non_exe(iparams0, '_bris', 0, 0.03, 0.15)
    #print imp.HighFrequency(iparams0, 'fixn', 0.0, '_bris', 0.03, 0.15)
    print imp.HighFrequency(iparams0, 's_oo', 0.0, '_bris', 0.03, 0.15)
    
