#include <iostream>
#include <fstream>
#include <vector>
#include "smesh.h"
#include "sfunction.h"

using namespace std;

bool CheckStream(istream& inputf, int& n, int& m, int& ncomment)
{
  istream input(inputf.rdbuf());
  input.seekg(0,ios::beg);

  n=0;
  string str;
  bool begincomm=false;
  getline(input,str); n++;
  if (!input.good()) {
    cerr << "ERROR: Wrong file format for input stream or no data!" << endl;
    return false;
  }
  ncomment=0;
  while (str[0]=='#'){
    clog<<str<<endl;
    getline(input,str);
    ncomment++;
  }
  clog<<"ncomment="<<ncomment<<endl;
  n++;
  stringstream oneline(str);
  m=0; double t;
  while (oneline){oneline>>t; m++;}
  m--;
  while (input){ getline(input,str); n++;}
  n--;

  clog << " Number of entries: "<< n <<endl;
  clog << " Number of columns: "<< m <<endl;

  inputf.seekg(0,ios::beg);
  if (begincomm) getline(inputf, str);
  if (!inputf){ cerr<<"Reopening didn't suceeded!"<<endl; return false;}
  return true;
}

bool ReadData(istream& input, mesh1D& om, function1D<dcomplex>& fi, int mdata, int ncomment)
{
  clog<<"Reading data with "<<om.size()<<" entries"<<endl;
  for (int i=0; i<ncomment; i++) input.ignore(5000,'\n');
  if (mdata<2) {cerr<<"Not enough columns in input Delta file!"<<endl; return false;}
  vector<double> data(mdata);
  int i=-1;
  while (input && ++i<om.size()){
    for (int j=0; j<mdata; j++) input>>data[j];
    input.ignore(500,'\n');
    if (i==0 || i==om.size()-1) clog<<"first/last frequency="<<data[0]<<endl;
    fi[i] = dcomplex(data[1],data[2]);
    om[i] = data[0];
  }
  input.clear();              // forget we hit the end of file
  input.seekg(0, ios::beg);   // move to the start of the file
  string str;
  getline(input, str);        // read comment
  return true;
}


double FindHighFrequencyExp(int Nf, const mesh1D& omi, const function1D<dcomplex>& fi)
{
  double ah;
  //  for (int b=0; b<fi.size_N(); b++){
  double S=0, Sx=0, Sy=0, Sxx=0, Sxy=0;
  for (int j=omi.size()-Nf; j<omi.size(); j++){
    double y = fi[j].imag()*omi[j];
    double x = omi[j];
    Sy += y;
    Sx += 1/(x*x);
    Sxx += 1/(x*x*x*x);
    Sxy += y/(x*x);
    S++;
    //clog<<"y="<<y<<" x="<<x<<endl;
  }
  double dd = S*Sxx-Sx*Sx;
  double a = (Sxx*Sy-Sx*Sxy)/dd;
  double bx = (S*Sxy-Sx*Sy)/dd;
  //    clog<<a<<" "<<bx<<endl;
  ah = -a;
  //  }
  if (fabs(ah-1.0)<1e-3) ah=1.0;
  return ah;
}

void InverseFourier(const function<dcomplex>& Giom, const mesh& iom, function<double>& Gtau, const mesh& tau, double ah, double beta)
{
  for (int t=0; t<tau.size(); t++){
    double sum=0;
    for (int n=0; n<iom.size(); n++)
      sum += cos(iom[n]*tau[t])*Giom[n].real() + sin(iom[n]*tau[t])*(Giom[n].imag()+ah/iom[n]);
    Gtau[t] = 2*sum/beta-0.5*ah;
  }

  // Correction 1/omega^2 (sometimes usefull)
  // corrected 16.12.2006
  double df = Giom.last().real()*iom.last()/M_PI;
  Gtau[0] += df;
  Gtau[Gtau.size()-1] -= df;  
  
}

int main(int argc, char *argv[], char *env[])
{
  string finp = "Gf.out";
  int Ntau = 1000;
  
  if (argc>1) Ntau = atoi(argv[1]);
  if (argc>2) finp = argv[2];
  
  clog<<"Number of time slices = "<<Ntau<<endl;
  
  ifstream input(finp.c_str());
  int n, m, ncomment;
  if (!CheckStream(input, n, m, ncomment)) { cerr<<"Something wrong in reading input Delta="<<finp<<endl; return 1;}
  
  mesh1D omi(n);
  function1D<dcomplex> Gw(n);

  if (!ReadData(input, omi, Gw, m, ncomment)) return false;
  omi.SetUp(0);
  double ah = FindHighFrequencyExp(40, omi, Gw);
  clog<<"ah="<<ah<<endl;
  double beta = M_PI/omi[0];
  mesh1D tau;
  tau.MakeEquidistantMesh(Ntau+1,0,beta);
  spline1D<double> Gt(tau.size());
  
  InverseFourier(Gw, omi, Gt, tau, ah, beta);
  double df0 = (Gt[1]-Gt[0])/(tau[1]-tau[0]);
  int nt = tau.size();
  double dfn = (Gt[nt-1]-Gt[nt-2])/(tau[nt-1]-tau[nt-2]);
  Gt.splineIt(tau,df0,dfn);

  //  DeltaFourier(Ntau, beta, tau, Delta, omi, Deltaw, ah);
  cout.precision(16);
  print(cout, tau, Gt, 25);
  return 0;
}
