#!/usr/bin/env python
import sys, re, os, glob, shutil
from os.path import getsize
from scipy import *

scratch = os.environ.get('SCRATCH')

ROOT = os.environ.get('WIEN_DMFT_ROOT')
if ROOT is not None:
    sys.path.append( ROOT )
else:
    print >> sys.stderr, "Environment variable WIEN_DMFT_ROOT must be set!"
    print "Environment variable WIEN_DMFT_ROOT must be set!"
    sys.exit(1)

import oca
import ctqmc

import numpy
if numpy.__version__ == '1.0.1':
    loadtxt = io.read_array
    savetxt = io.write_array
import copy


class Params(object):
    """Class to store various control paramers which control the main flow of the program
       The class is constructed such that it can handle two types of parameters:
         - starting parameters which are read only at the instantiation
         - on fly parameters which can be refreshed (reread from the file) at any time
           by a member function refresh

       Input:
          f_sparams    -- filename of the parameters read only at the beginning
          sparams      -- a dictionary of pairs {name of the variable: default value} which are looked for in the f_sparams file
          f_params     -- filename of the parameters which can be refreshed several times by refresh member function
          params       -- a dictionary of {variable: default value} pairs which can be updated at runtime
       Output:
          None
       Variables which become members of the class are those listed in params and sparams dictionaries
    """
    def __init__(self, f_params, params, f_sparams=None):

        self.p ={}
        self.p.update(params) # default values from the __main__

        # remember params and its data file
        self.f_params = f_params
        
        # executes the parameters at startup
        if f_sparams is not None:
            execfile(f_sparams)
            sp = locals()      # stores parameters
            del sp['self']     # deletes automatic locals
            self.p.update(sp)  # new or updates variables
        
    def refresh(self, fh_info):

        # executes the parameters file
        execfile(self.f_params)
        sp = locals()   # stores locals
        del sp['self']  # deletes automatics

        # prints what was changed
        for c in sp.keys():
            if (c not in self.p.keys()) or (sp[c] != self.p[c]):
                print >>fh_info, '********------ parameter change ------***********'
                print >>fh_info, c, "=", sp[c]

        # updates the dictonary
        self.p.update(sp)
        
    def __getitem__(self, x):
        return self.p[x]

    def keys(self):
        return self.p.keys()



class Struct:
    def __init__(self, filename, fh_info=sys.stdout):
        """ Reads Win2K structure file"""
        print >> fh_info, '***** Reading Structure file *****'
        fh = open(filename, 'r')

        self.title = fh.next()
        line = fh.next()
        lattic = line[:4]
        self.nat = int(line[27:30])
        print >> fh_info, 'Number of sorts, nat=', self.nat
        
        fh.next()
        line = fh.next()
        self.angle=zeros(3,dtype=float)

        (self.a,self.b,self.c,self.angle[0],self.angle[1],self.angle[2]) = map(float, [line[:10],line[10:20],line[20:30],line[30:40],line[40:50],line[50:60]])
        #(self.a,self.b,self.c,self.angle[0],self.angle[1],self.angle[2]) = map(float, line.split())
        for i in range(3):
            if abs(self.angle[i])<1e-5: self.angle[i]=90.
        print >> fh_info, 'unit vectors (a,b,c) are ', self.a, self.b, self.c
        print >> fh_info, 'angles betwen unit vectors are', self.angle

        print >> fh_info, '---- over all types and atoms ----'
        self.iatnr=[]
        self.pos=[]
        self.mult=[]
        self.isplit=[]
        self.aname=[]
        self.jrj=[]
        self.r0=[]
        self.rmt=[]
        self.Znuc=[]
        self.RotLoc=[]
        for iat in range(self.nat):
            print >> fh_info, '---------- atom type', iat, '------------'
            line = fh.next()
            self.iatnr.append( int(line[4:8]) )
            print >> fh_info, 'iatnr=', self.iatnr[-1]
            tpos = map(float, [line[12:22], line[25:35],line[38:48]])
            self.pos.append(tpos)
            print >> fh_info, 'Position inside the unit cell, pos=', tpos
            
            line = fh.next()
            self.mult.append( int(line[15:17]) )
            print >> fh_info, 'Number of atoms of this type, mult=', self.mult[-1]
            self.isplit.append( int(line[34:36]) )
            print >> fh_info, 'Symmetry of the atom, isplit=', self.isplit[-1]

            if self.mult[-1]>1: print >> fh_info, '   *** Equivalent atoms ***'
            
            for mu in range(self.mult[-1]-1):
                line = fh.next()
                self.iatnr.append( int(line[4:8]) )
                print >> fh_info, '    iatnr=', self.iatnr[-1],
                self.pos.append( map(float, [line[12:22], line[25:35],line[38:48]]) )
                print >> fh_info, '    pos=', self.pos[-1]
                
            line = fh.next()
            self.aname.append( line[:10] )
            self.jrj.append( int(line[15:20]) )
            self.r0.append( float(line[25:35]) )
            self.rmt.append( float(line[40:50]) )
            self.Znuc.append( float(line[55:65]) )
            print >> fh_info, '--------- Information about the current type -------------'
            print >> fh_info, 'aname=', self.aname[-1], 'Znuc=', self.Znuc[-1], 'Rmt=', self.rmt[-1]
            rt=[]
            for i in range(3):
                line = fh.next()
                rt.append( map(float, [line[20:30], line[30:40],line[40:50]]) )
            self.RotLoc.append(rt)
            print >> fh_info, 'RotLoc=', self.RotLoc[-1]
            print >> fh_info, '-----------------------------------------------------------'

        self.aZnuc=[]
        for iat in range(self.nat):
            for mu in range(self.mult[iat]):
                self.aZnuc.append(self.Znuc[iat])
        
        print >> fh_info, 'Z=', self.aZnuc
        print >> fh_info
        
        
        
        
def ParsIndmft(filename):
    """Read case.indmft file"""
    def union(data):
        " Takes a union of array or list"
        c = []
        for d in data:
            if d not in c:
                c.append(d)
        return c
    
    fh_indmft = open(filename, 'r')
    data = fh_indmft.readlines()

    matsubara = int(data[1].split()[0])
    
    natom = int(data[2].split()[0])
    iatom={}
    it=3
    for iat in range(natom):
        (iat,nl,locrot) = map(int, data[it].split()[:3])
        loro = locrot % 3
        shft = locrot / 3
        #print 'iat,nl,locrot,loro,shft=', iat, nl, locrot, loro, shft
        it += 1
        for il in range(nl):
            (l, qsplit, icix) = map(int, data[it].split()[:3])
            iatom[icix] = iat
            #print 'l,qsplit,icix=', l, qsplit, icix, iatom
            it += 1
        it += loro + shft
    it+=1 # comment
    it+=1 # mode,EF
    it+=1 # comment
    
    (ncix,maxdim,maxsize) = map(int,data[it].split()[:3])
    it+=1

    Sigind=[]
    csize=[]
    CF=[]
    for icix in range(ncix):
        (wicix, wndim, size) = map(int, data[it].split()[:3]); it+=1
        csize.append(size)
        it+=1 # comment: Independent components
        legend = data[it]; it+=1
        it+=1 # comment: Sigind follows
        wSigind=[]
        for i in range(wndim):
            wSigind.append( map(int, data[it].split()[:wndim]) )
            it+=1
        Sigind.append(wSigind)
        it+=1 # comment: Transformation matrix
        tCF=[]
        for i in range(wndim):
            vd = map(float, data[it].split()[:2*wndim])
            #print array(vd[0::2])
            #print array(vd[1::2])*1j
            tCF.append( array(vd[0::2])+array(vd[1::2])*1j )
            it+=1
        CF.append(tCF)
    CF = array(CF)
        
    #print csize
    #print Sigind

    uimpurities={}
    impurities=[]
    for icix in range(ncix):
        comps = sort(union(array(Sigind[icix]).flatten())).tolist()
        comps.remove(0)
        
        impurities.append(comps)
        unique=True
        for icx,cmps in uimpurities.items():
            diff = [val for val in cmps[0] if val not in comps]
            #print ':', comps, cmps[0], diff, (not diff)
            if not diff:
                unique=False
                equivalent_cix = icx
                break
        if unique:
            uimpurities[icix+1]=[comps, [icix+1] ]
        else:
            print 'Found non-unique cix', equivalent_cix, icix
            uimpurities[equivalent_cix][1].append(icix+1)

    wuimpurities={}
    for iw,w in uimpurities.items():
        wuimpurities[ tuple(w[1]) ] = w[0]
    #print 'wuimpurities=', wuimpurities    
    uimpurities = wuimpurities

    #Impq=[]
    #for atms,cals in uimpurities.items():
    #    icix = atms[0]-1
    #    t_Impq = [Sigind[icix][l][l] for l in range(len(Sigind[icix]))]
    #    Impq.append( array(t_Impq) - min(t_Impq) )
    
    return (impurities, uimpurities, iatom, Sigind, CF, matsubara)


def ReadSig_inp(signame, acols, fh_info=sys.stdout):
    """ Reading self-energy """
    fh_sig = open(signame, 'r')
    # First s_oo and Edc
    m = re.search(r'(s_oo\s*=\s*\[.*\])', fh_sig.next())
    if m is not None:
        exec(m.group(1))
    m = re.search(r'(Edc\s*=\s*\[.*\])', fh_sig.next())
    if m is not None:
        exec(m.group(1))
    fh_sig.close()
    print >> fh_info, 's_oo=', s_oo
    print >> fh_info, 'Edc=', Edc
    
    if len(s_oo)<len(acols) or len(Edc)<len(acols):
        print 'ERROR: not enough s_oo components or Edc components in the input self-energy', signame
        print 'Boling out......'
        sys.exit(1)
    
    # Reading the rest of the data from self-energy file
    sigdata = loadtxt(signame).transpose()

    return (s_oo, Edc, sigdata)


def create_log_mesh(sigdata, nom, ntail_):
    """Creates logarithmic mesh on Matsubara axis
       Takes first istart points from mesh om and
       the rest of om mesh is replaced by ntail poinst
       redistribued logarithmically.
       Input:
           om      -- original long mesh
           istart  -- first istart points unchanged
           ntail   -- tail replaced by ntail points only
       Output:
           som             -- smaller mesh created from big om mesh
           sSig[nom,nc]    -- Sig on small mesh
       Also computed but not returned:
           ind_om  -- index array which conatins index to
                      kept Matsubara points
    """
    om = sigdata[0]
    
    istart = min(nom, len(om))
    ntail = min(ntail_, len(om)-istart)
    
    
    istart = min(nom,len(om))
    ntail = min(ntail, len(om)-istart)
    
    ind_om=[]
    alpha = log((len(om)-1.)/istart)/(ntail-1.)
    for i in range(istart):
        ind_om.append(i)
    for i in range(ntail):
        t = int(istart*exp(alpha*i)+0.5)
        if (t != ind_om[-1]):
            ind_om.append(t)

    ssigdata = zeros( (shape(sigdata)[0], len(ind_om)), dtype=float )
    for i in range(len(ind_om)):
        ssigdata[:,i] = sigdata[:,ind_om[i]]

    return ssigdata


def PrepareAllSigmas(impurities, sigdata, s_oo, Edc, fh_info=sys.stdout):
    """ Creates the input self-energies file for k-sum"""
    for icix,imp in enumerate(impurities):  # over all imp.problems, even those that are equivalent
        wdata=[]
        wdata.append(sigdata[0])
        sinf=[500.]  # adding sigma_oo

        #print 'Processing impurity', icix, imp
            
        for col in imp:

            #print 'taking', 2*col-1, 2*col, col-1, 'and adding', s_oo[col-1], Edc[col-1]
            #sigdata[2*col-1] += s_oo[col-1]-Edc[col-1]  # adding s_oo-Edc  !!! FOUND A BUG!!!
            real_part = sigdata[2*col-1] + s_oo[col-1]-Edc[col-1] # adding s_oo-Edc ! corrected
            
            wdata.append(real_part)                     # corresponds to real-part of this imp. problem
            wdata.append(sigdata[2*col])                # corresponds to imag-part of this imp. problem
            sinf.append(s_oo[col-1]-Edc[col-1]); sinf.append(0) # s_oo also needed in k-sum for Eimp

        fname = 'sig.inp'+str(icix+1)                   # this is the name of the self-energy for ksum
        wdata = array(wdata).transpose()
        
        wdata = vstack((wdata, sinf))                   # adding s_oo as the last line
        savetxt(fname, wdata)                           # writting to the disc the self-energy
        
        print >> fh_info, 'shape(wdata)=', shape(wdata)
        print >> fh_info, 'sinf=', sinf



def FindLDAChemicalPotential(case, updn):
    Ry2eV = 13.6058
    # Looking for the LDA chemical potential

    if  os.path.isfile(case+".scf2"):
        fscf = open(case+".scf2", 'r')
    else:
        fscf = open(case+".scf2"+updn, 'r')
    
    lines = fscf.readlines()
    for line in lines:
        if re.match(r':FER', line) is not None:
            EF_LDA = float(line[38:])*Ry2eV
            print 'EF_LDA=', EF_LDA
            break
    ## The previous DMFT chemical potential
    #if  os.path.isfile('EF.dat'):
    #    fmu = open('EF.dat','r')
    #    EF_C = float(fmu.next())
    #    print 'Found DMFT-mu=', EF_C
    return (EF_LDA)




if __name__ == '__main__':
    # -------------- Parameters which are set through input files sparams.dat and params.dat -----------
    # Default values of some parameters
    params = {'runIMP'        : True,
              'runGF'         : True,
              'UpdateAtom'    : False,
              'DCs'           : 'fixn',
              'DCs'           : 'S0',
              #'DCs'          : 'default',
              #'DCs'          : 'fixEimp',
              'Niter'         : 1,
              'max_iterations': 1000,
              'finish'        : 100,
              'da'            : 0.0,
              'wbroad'        : 0.0,
              'kbroad'        : 0.0,
              'recompute_mu'  : True,
              'solver'        : 'OCA',
              #'solver'       : 'CTQMC',
              'com'           : 0, # when computing density on imaginary axis, we can subtract Sigma_oo (com=0) or Sigma(0) (com=1)
              'orthog'        :False,
              'ntail'         :30,
              'mix_delta'     :1.0,
              'magn_ind'      : None,
              }

    extn=''
    
    p = Params('params.dat', params)

    # Info file
    fh_info = open('dmft_info.out','w')
    fh_pinfo = open('info.iterate', 'a')
    fh_pinfo.write('%3s.%3s %12s %12s %12s %12s\n' % ('#','#','mu','Eimp','Edc','nf'))
    fh_pinfo.flush()
    
    if len(glob.glob('ksum_info.out'))!=0 : os.remove('ksum_info.out')
    
    # For parallel execution of the impurity solver
    mpi_prefix=''
    if len(glob.glob('mpi_prefix.dat'))!=0:
        mpi_prefix = open('mpi_prefix.dat', 'r').next().strip()

    # Reading parameters from params.dat
    p.refresh(fh_info)
    
    # Determines 'case'
    pwd = os.getcwd()
    case = os.path.split(pwd)[1]
    if len(glob.glob(case+'.struct'))==0: # the directory name is not case. This happens on node when submitting
        case = os.path.splitext(glob.glob("*.struct")[0])[0] # finds file xxx.struct and sets case to xxx
    
    # Reads structure file
    struct = Struct(case+'.struct', fh_info)

    
    # Determines the number of impurity problems
    (impurities, uimpurities, iatom, Sigind, CF, matsubara) = ParsIndmft(case+'.indmf0')

    #magn_ind=[1,1,0,0]    
    magn_ind=p['magn_ind']
    if magn_ind is not None:
        updn = ['-up', '-dn']
    
    if len(updn)>1:
        r_impurities = [copy.deepcopy(impurities), copy.deepcopy(impurities)]
        r_uimpurities = [copy.deepcopy(uimpurities), copy.deepcopy(uimpurities)]
        
        maxcol = max(array(uimpurities.values()).flatten())
        
        for icix in range(len(r_impurities[0])):
            mi = magn_ind[icix]
            for c in range(len(r_impurities[0][icix])):
                r_impurities[mi][icix][c] += maxcol
                
        for icix in r_uimpurities[0].keys():
            mi = magn_ind[0]
            for c in range(len(r_uimpurities[0][icix])):
                r_uimpurities[mi][icix][c] += maxcol

        for icix in range(len(Sigind)):
            (n1, n2) = shape(Sigind[icix])
            if (n1!=n2): print 'ERROR: Sigind not a square matrix'
            tSigind= zeros((2*n1,2*n2),dtype=int)
            #tSigind[:n1,:n2] = Sigind[icix]
            for ip in range(n1):
                for iq in range(n2):
                    if Sigind[icix][ip][iq]>0:
                        mi = magn_ind[icix]
                        tSigind[n1+ip,n2+iq] = Sigind[icix][ip][iq] + maxcol*mi
                        tSigind[   ip,   iq] = Sigind[icix][ip][iq] + maxcol*(1-mi)
                        
            Sigind[icix] = tSigind.tolist()

        CFN = [[] for i in range(len(CF))]
        for icix in range(len(CF)):
            CFN[icix] = zeros((2*n1,2*n2), dtype=complex)
            CFN[icix][:n1,:n2] = CF[icix]
            CFN[icix][n1:,n2:] = CF[icix]
        CF = CFN

    else:
        r_impurities = [impurities]
        r_uimpurities = [uimpurities]

    r_acols = []
    for uimpurities in r_uimpurities:        
        r_acols += array(uimpurities.values()).flatten().tolist()
    r_acols.sort()

    print >> fh_info, 'maxcol=', maxcol
    print >> fh_info, 'r_impurities=', r_impurities
    print >> fh_info, 'r_uimpurities=', r_uimpurities
    print >> fh_info, 'r_acols=', r_acols
    print >> fh_info, 'iatom=', iatom
    print >> fh_info, 'Sigind=', Sigind

    # Number of independent columns which we need in sig.inp file
    #acols = sort(array(uimpurities.values()).flatten())
    #
    #print >> fh_info, 'impurities=', impurities
    #print >> fh_info, 'uimpurities=', uimpurities
    #print >> fh_info, 'acols=', acols
    #print >> fh_info, 'iatom=', iatom
    impurities = None
    uimpuritis = None
    acols = None
    
    # atoms in each atom-type
    atms = uimpurities.keys()
    atms.sort()

    # nuclear charge of correlated atoms
    Znuc = [int(struct.aZnuc[iatom[atm[0]]-1]) for atm in atms]
    print >> fh_info, 'atoms=', [iatom[atm[0]] for atm in atms]
    print >> fh_info, 'Znucs=', Znuc

    # Solvers initialized
    if p['solver'] in ['OCA', 'NCA']:
        if matsubara == True:
            print 'ERROR: Can not run OCA/NCA on imaginary axis! Change case.idmf0:matsubara or params.dat:solver'
        solver = [oca.IMP_OCA(Znuc[i], 'imp.'+str(i), p['iparams'+str(i)], Sigind[i], CF[i]) for i in range(len(atms))]
    elif p['solver'] in ['CTQMC']:
        if matsubara == False:
            print 'ERROR: Can not run CTQMC on real axis! Change case.idmf0:matsubara or params.dat:solver'
        solver = [ctqmc.IMP_CTQMC(Znuc[i], 'imp.'+str(i), p['iparams'+str(i)], Sigind[i], CF[i]) for i in range(len(atms))]
    else:
        print 'ERROR: No solver specified! Bolling out.'

    shutil.copy2(ROOT+'/dmft', '.')
    
    print >> fh_info, '--------- Preparing GF calculation ---------'
    cmd = ROOT+'/x_dmft.py dmft2 '+ updn[0] +' >> ksum_info.out 2>&1'
    print >> fh_info, cmd
    stdin, stdout, stderr = os.popen3(cmd)
    print >> fh_info, stdout.read(), stderr.read()



    #print '------- Recompute eigenvectors ------' # They might be on scratch and we need to recompute them
    #lapw_end = 'lapw1'
    #so = ''
    #if os.path.isfile(scratch+"/"+case+".vectorso") and os.path.getsize(scratch+"/"+case+".vectorso")>0 :
    #    lapwend = 'lapwso'
    #    so = '-so'
    #cmd = 'run_lapw '+so+' -s lapw0 -e '+lapwend
    #print cmd
    #stdin, stdout, stderr = os.popen3(cmd)
    #print stdout.read(), stderr.read()
    #if getsize('lapw0.error') !=0:
    #    print 'ERROR in lapw0', open('lapw0.error','r').next()
    #    exit(1)
    #if getsize('lapw1.error') !=0:
    #    print 'ERROR in lapw1', open('lapw1.error','r').next()
    #    exit(1)
    #if getsize('lapwso.error') !=0:
    #    print 'ERROR in lapwso', open('lapwso.error','r').next()
    #    exit(1)
    
    
    shutil.copy2('sig.inp', 'sig.inp.0')

    for itt in range(p['max_iterations']):
        
        extn = str(itt+1)  # Extension of output files
        p.refresh(fh_info)
        
        # Reading self-energy    
        (s_oo, Edc, sigdata) = ReadSig_inp('sig.inp', r_acols, fh_info)
        nom = shape(sigdata)[1]

        #####
        if matsubara:
            lsigdata = copy.deepcopy(sigdata)
            lom = lsigdata[0]
            sigdata = create_log_mesh(sigdata, p['iparams0']['nom'][0], p['ntail'])
                                      # nom stands for the full long mesh
            snom = shape(sigdata)[1]  # snom stands for the small logarithim mesh
            print >> fh_info, "Small mesh of size", shape(sigdata)[1], "frequency points created"
        else:
            snom = nom
            lom = sigdata[0]
        #####



        if p['recompute_mu']:

            for iupd,upd in enumerate(updn):
                # Creates the input self-energies file for k-sum
                PrepareAllSigmas(r_impurities[iupd], sigdata, s_oo, Edc, fh_info)
                fh_info.flush()
            
                print >> fh_info, '\n--------- Computing eigenvalues '+upd+'---------'
                print >> fh_info, 'Running ---- ksum ----'
                cmd = mpi_prefix+' ./dmft dmft0.def >> ksum_info.out 2>&1'
                print >> fh_info, cmd
                fh_info.flush()
                stdin, stdout, stderr = os.popen3(cmd)
                print >> fh_info, stdout.read(), stderr.read()
                for fe in glob.glob('dmft0.error*'):
                    if getsize(fe) !=0:
                        print 'ERROR in dmft0', open('dmft0.error','r').next()
                        exit(1)
                shutil.move('eigvals.dat', 'eigvals.dat'+upd)

            print >> fh_info, '--------- Determine the chemical potential ---------'
            fh_info.flush()
            cmd = ROOT+'/x_dmft.py dmft1 '+updn[0]
            print >> fh_info, '\nRunning ---- x_dmft.py ----'
            print >> fh_info, cmd
            stdin, stdout, stderr = os.popen3(cmd)
            print >> fh_info, stdout.read(), stderr.read()
            fh_info.flush()
            
        # Computing Green's function
        if p['runGF']:
            for iupd,upd in enumerate(updn):
                # Creates the input self-energies file for k-sum
                PrepareAllSigmas(r_impurities[iupd], sigdata, s_oo, Edc, fh_info)
                fh_info.flush()

                shutil.copy2('sig.inp1', 'brisi_sig.inp1'+upd)
                shutil.copy2('sig.inp2', 'brisi_sig.inp2'+upd)
                shutil.copy2('sig.inp3', 'brisi_sig.inp3'+upd)
                shutil.copy2('sig.inp4', 'brisi_sig.inp4'+upd)
                
                print >> fh_info, '--------- Preparing indmft2.def ---------'
                cmd = ROOT+'/x_dmft.py dmft2 '+upd
                
                print >> fh_info, cmd
                fh_info.flush()
                stdin, stdout, stderr = os.popen3(cmd)
                print >> fh_info, stdout.read(), stderr.read()

                print >> fh_info, 'Running ---- ksum -----'
                cmd = mpi_prefix + ' ./dmft dmft2.def >> ksum_info.out 2>&1 '
                print >> fh_info, cmd
                fh_info.flush()
                stdin, stdout, stderr = os.popen3(cmd)
                print >> fh_info, stdout.read(), stderr.read()
                
                for fe in glob.glob('dmft2.error*'):
                    if getsize(fe) !=0:
                        print 'ERROR in dmft2', open('dmft2.error','r').next()
                        exit(1)
                
                shutil.copy2(case+'.cdos', case+'.cdos'+upd+'.'+extn)
                shutil.copy2(case+'.gc1', case+'.gc1'+upd+'.'+extn)
                shutil.copy2(case+'.dlt1', case+'.dlt1'+upd+'.'+extn)


                shutil.copy2(case+'.gc1', 'brisi_gc1'+upd)
                shutil.copy2(case+'.gc2', 'brisi_gc2'+upd)
                shutil.copy2(case+'.gc3', 'brisi_gc3'+upd)
                shutil.copy2(case+'.gc4', 'brisi_gc4'+upd)
                
                if (upd):
                    ncix = len(r_impurities[iupd])
                    for icix in range(ncix):
                        shutil.copy2(case+'.Eimp'+str(icix+1), case+'.Eimp'+str(icix+1)+upd)
                        shutil.copy2(case+'.dlt'+str(icix+1), case+'.dlt'+str(icix+1)+upd)


        C_sinf = zeros(len(r_acols), dtype=float)
        C_Edc  = zeros(len(r_acols), dtype=float)
        C_nf_IMP = zeros(len(atms), dtype=float)
        C_Sig = zeros((len(r_acols)*2,nom), dtype=float)

        
        # ------------ Impurity solver -------------------
        if p['runIMP']:

            for iat,atm in enumerate(atms): # Over all inequivalent impurity problems

                r_cols = r_uimpurities[0][atm] + r_uimpurities[1][atm]
                ncols = len(r_cols)/2
                aver = [zeros((snom,1+2*ncols), dtype=float), zeros((snom,1+2*ncols), dtype=float)]
                Eimp = [zeros(ncols, dtype=float), zeros(ncols, dtype=float)]
                
                for iupd,upd in enumerate(updn):
                    cols = r_uimpurities[iupd][atm]
                    print >> fh_info, 'Creating ---- input Delta -----'
                    print 'iat=', iat, 'atm=', atm, 'cols=', cols
                    # averagin over all equivalent Delta's
                    for at in atm:
                        mi = magn_ind[at-1]*iupd + (1-magn_ind[at-1])*(1-iupd)
                        print 'iat=', iat, 'iupd', iupd, 'at=', at, 'magn_ind', mi
                        
                        fname = case+'.dlt'+str(at)+upd
                        aver[mi] += loadtxt(fname)
                        execfile(case+'.Eimp'+str(at)+upd)
                        Eimp[mi] += array(Ed)
                        
                aver = array(aver)/len(atm)
                Eimp = hstack(tuple(Eimp))/len(atm)
                
                aver = hstack((aver[0][:,:],aver[1][:,1:])) # frequency appears in both, so remove it in the second part
                
                if p['mix_delta']!=1.0:
                    print 'mixing delta'
                    deltamix = 'delta.'+str(iat)
                    Emix = 'Eimp.'+str(iat)
                    if len(glob.glob(deltamix))!=0:
                        aver_old = loadtxt(deltamix)
                        Eimp_old = loadtxt(Emix)
                        if abs(sum(array(aver_old[:,0])-array(aver[:,0])))<1e-5: # the same mesh
                            print 'mesh is the same, mixing'
                            waver = array(aver)*p['mix_delta'] + array(aver_old)*(1.-p['mix_delta'])
                            wEimp = array(Eimp)*p['mix_delta'] + array(Eimp_old)*(1.-p['mix_delta'])
                            aver = waver
                            Eimp = wEimp
                    savetxt(deltamix, aver)
                    savetxt(Emix, Eimp)

                
                Delta = aver[:,1::2] + aver[:,2::2]*1j
                om = aver[:,0]

                print 'r_cols=', r_cols
                # subtracting Edc, because the ksum used (Hk+s_oo-Edc)
                ws_oo = [s_oo[c-1] for c in r_cols]
                wEdc = [Edc[c-1] for c in r_cols]
                Eimp -= array(wEdc)
                print >> fh_info, 'Eimp=', Eimp, 'Edc=', wEdc
                
                #print 'Running ----- impurity solver -----'
                print >> fh_info, 'Running ----- impurity solver -----'
                iprms = 'iparams'+str(iat)
                p[iprms]['Ed'] = [Eimp.tolist(), "# Impurity levels"]
                fh_info.flush()

                if p['solver'] in ['OCA','NCA']:
                    solver[iat]._exe(p[iprms], mpi_prefix, om, Delta, p['wbroad'], extn, Eimp, p['UpdateAtom'])
                elif p['solver'] in ['CTQMC']:
                    solver[iat]._exe(p[iprms], mpi_prefix, lom, om, Delta, extn, Eimp, p['UpdateAtom'])
                else:
                    print 'ERROR: Impurity solver not defined!'
                fh_info.flush()

                print >> fh_info, 'Taking care of high frequency'
                (lom, t_Sig, t_sinf, t_Edc, C_nf_IMP[iat]) = solver[iat].HighFrequency(p[iprms], p['DCs'], p['da'], extn, p['wbroad'], p['kbroad'])
                fh_info.flush()

                if len(glob.glob('EF.dat'))!=0:
                    dat=open('EF.dat', 'r').readlines()
                    EF = float(dat[0].split()[0])
                else:
                    if len(updn)>1:
                        EF_LDA = FindLDAChemicalPotential(case, updn[0][1:])
                    else:
                        EF_LDA = FindLDAChemicalPotential(case, '')
                    EF=EF_LDA
                    
                print >> fh_pinfo, '%3d.%3d %12.6f %12.6f %12.6f %12.6f' % (itt,0,EF,Eimp[0],wEdc[0], C_nf_IMP[iat])
                fh_pinfo.flush()
                
                
                # Combining all impurity problems into combined output
                for j in range(len(t_sinf)):
                    c = r_cols[j]-1
                    print 'j,c=', j, c
                    C_sinf[c] = t_sinf[j]
                    C_Edc[c] = t_Edc[j]
                    C_Sig[2*c,:] = real(t_Sig[j])
                    C_Sig[2*c+1,:] = imag(t_Sig[j])
            
            
            print >> fh_info, 'sinf=', C_sinf
            print >> fh_info, 'Edc=', C_Edc
            print >> fh_info, 'nf_IMP=', C_nf_IMP
            
            
            filesig = 'sig.inp.'+extn
            fs = open(filesig, 'w')
            print >> fs, '# s_oo=', C_sinf.tolist()
            print >> fs, '# Edc=', C_Edc.tolist()
            savetxt(fs, vstack((lom,C_Sig)).transpose())
            fs.close()
            
            shutil.copy2('sig.inp.'+extn, 'sig.inp')
        print >> fh_info, '------- Iteration', itt, 'finished ---------'
        fh_info.flush()
        
